/*
   Texture Mapping Demo.
   Created by the Programming Ace.
   www.UltimateGameProgramming.com

   The purpose of this demo is to load a texture into Direct3D.
   A texture is an image that is displayed on an object (square, character model, etc)
   to give that object a new dimension of realism.  In this demo we will apply a
   texture on a square.  In order to correctly apply a texture
   you must set a texture coordinate for each vertex.  For our square it could look like this...

   Vertex Polygon_Data[4] =
   {
      //    x     y      z     1    2
		   {-0.5f, 0.5f,  0.0f, 0.0, 0.0},
         {0.5f,  0.5f,  0.0f, 1.0, 0.0},
	      {-0.5f, -0.5f, 0.0f, 0.0, 1.0},
         {0.5f,  -0.5f, 0.0f, 1.0, 1.0}
	};

   The x, y, z are position points.  1 and 2 are the texture coordinates.  A tex coord
   can be either 0.0 or 1.0.  Lets take a look at a image below...

   (0,0) _____ (1,0)
        |     |
        |     |
        |     |
   (0,1)|_____|(1,1)


   Here you can see which texture coords go with which corner of the image.
   0,0 is the upper left of the image, 1,0 is the upper right, etc.
   Now if you look at the information for our square above you will see that
   the first point if the upper left corner of the square so it gets the matching
   texture coordinate and so on for each point.  If you ever done texture mapping
   in OpenGL you will notice that the tex coords are different in Direct3D.

   To load a image into Direct3D you will take these steps...

   Step 1: Add D3DFVD_TEX1 to the vertex structure and define a texture object like so...

      #define D3DFVF_D3DVertex (D3DFVF_XYZ|D3DFVF_TEX1)
      LPDIRECT3DTEXTURE9 Texture = NULL;


   Step 2: Add two floats to the vertex struct for the texture coords like so...

      // Vertex stucture.
      struct Vertex
      {
         FLOAT x, y, z;    // x, y, and z coordinates.
         FLOAT tu, tv;     // Texture coordinates.
      };


   Step 3: Load the texture image into memory (can use a function like D3DXCreateTextureFromFile())...

      if(D3DXCreateTextureFromFile(D3D_Device, "ugp.tga", &Texture) != D3D_OK)
         return false;


   Step 4: When you set the vertex buffer you must remember to define the tex coords like so...

      Vertex Polygon_Data[4] =
      {
		   {-0.5f, 0.5f, 0.0f, 0.0, 0.0},
         {0.5f, 0.5f, 0.0f, 1.0, 0.0},
	      {-0.5f, -0.5f, 0.0f, 0.0, 1.0},
         {0.5f, -0.5f, 0.0f, 1.0, 1.0}
	   };


   Step 5: Set the texture before drawing the object in the render using SetTexture() like so...
      
      D3D_Device->SetTexture(0, Texture);


   Step 6: Release the texture object when you are done...
      
      Texture->Release();
      Texture = NULL;

   Now on to the demo...

*/


// Demo: Texture Mapping.


#define WIN32_LEAN_AND_MEAN
#define VC_LEANMEAN

#include<d3dx9.h>


// Our vertex stucture define.
#define D3DFVF_D3DVertex (D3DFVF_XYZ | D3DFVF_TEX1)


// Function Prototypes...
bool InitializeDirect3D(HWND hwnd, bool fullscreen);// Initialize Direct3D.
bool InitializeObject();                        // Initialize object we are displaying.
void RenderScene();                             // Render the 3D scene.
void ShutdownDirect3D();                        // Shutdown Direct3D.


LPDIRECT3D9 Direct3D_Object = NULL;             // Direct3D object.
LPDIRECT3DDEVICE9 D3D_Device = NULL;            // Device object.
LPDIRECT3DVERTEXBUFFER9 Vertex_Buffer = NULL;   // Vertex buffer for hold our main object.
LPDIRECT3DTEXTURE9 Texture = NULL;              // Texture object to hold texture image.


// Vertex stucture.
struct Vertex
{
   FLOAT x, y, z;                               // x, y, and z coordinates.
   FLOAT tu, tv;                                // Texture coordinates.
};


LRESULT CALLBACK WndProc(HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam)
{
    switch(message)
       {
		   case WM_KEYUP:
            // If the user presses the escape key then exit the application.
            if(wParam == VK_ESCAPE)
               PostQuitMessage(0);
            break;


         case WM_DESTROY:
         case WM_CLOSE:
            PostQuitMessage(0);
            break;


            default:
               break;
       }

    return DefWindowProc( hwnd, message, wParam, lParam );
}


int WINAPI WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance, LPSTR lpCmdLine, int nShowCmd)
{
   MSG msg;                                           // Message object.
   HWND hwnd;                                         // Handle to the window.
   WNDCLASSEX windowClass;                            // Window class object.
   bool done = false;                                 // True if running, false to exit.

   // This is the Window class.
   windowClass.cbSize = sizeof(WNDCLASSEX);           // size of the WNDCLASSEX structure.
   windowClass.style = CS_HREDRAW | CS_VREDRAW;       // style of the window.
   windowClass.lpfnWndProc = WndProc;                 // Address to the windows procedure.
   windowClass.cbClsExtra = 0;                        // Extra class information.
   windowClass.cbWndExtra = 0;                        // Extra window information.
   windowClass.hInstance = hInstance;                 // Handle of application Instance.
   windowClass.hIcon = LoadIcon(NULL, IDI_APPLICATION);// Window Icon.
   windowClass.hCursor = LoadCursor(NULL, IDC_ARROW); // Load mouse cursor.
   windowClass.hbrBackground = NULL;                  // Background color.
   windowClass.lpszMenuName = NULL;                   // Menu.
   windowClass.lpszClassName = "UGPClass";            // Name of the window class.
   windowClass.hIconSm = LoadIcon(NULL, IDI_APPLICATION);// Minimized window icon.

   // You must register you class with Windows.
   if(!RegisterClassEx(&windowClass)) return 0;

   // After your class has been registered then you are ready to create your window.
   hwnd = CreateWindowEx(NULL,                                       // The extended style.
                         "UGPClass",                                 // Window class.
                         "Direct3D Texture Mapping - by The Programming Ace",// Window name.
                          WS_OVERLAPPEDWINDOW | WS_VISIBLE |         // Window style.
                          WS_SYSMENU |WS_CLIPCHILDREN |              // Window style.
                          WS_CLIPSIBLINGS,                           // Window style.
                          100, 100,                                  // X, Y coords.
                          640, 480,                                  // Window size.
                          NULL,                                      // Handle to parent window.
                          NULL,                                      // Menu.
                          hInstance,                                 // Handle to app instance.
                          NULL);                                     // Pointer to window.

   // If there was an error with creating the window, then close the program.
   if(!hwnd) return 0;

   ShowWindow(hwnd, SW_SHOW);    // Show the window.
   UpdateWindow(hwnd);           // Update its display.

   done = false;                 // false = run program, true means stop.

   // Initialize Direct3D.  If fails then we don't want to run the application.
   if(!InitializeDirect3D(hwnd, false)) done = true;

   // Application loop.
   while(!done)
      {
         if(PeekMessage(&msg, NULL, NULL, NULL, PM_REMOVE))
            {
               // If a quit message is received then stop rendering and quit the app.
               if(msg.message == WM_QUIT)
                  {
                     done = true;
                  }

               TranslateMessage(&msg);
               DispatchMessage(&msg);
            }
         else
            {
               RenderScene();
            }
      }


    // Here we shutdown Direct 3D.
	ShutdownDirect3D();

   // Here we unregister the window class with the OS.
   UnregisterClass("UGPClass", windowClass.hInstance);

   return (int)msg.wParam;
}


bool InitializeDirect3D(HWND hwnd, bool fullscreen)
{
   // This object will allow us to set the display mode of the screen.
   D3DDISPLAYMODE DisplayMode;

   // This will allow us to set the parameters of the screen.
   D3DPRESENT_PARAMETERS Present_Parameters;

   // This is used to get the capabilities of the hardware.
   D3DCAPS9 D3DCaps;

   // It is always a good idea to clear out memory in object although not necessary.
   ZeroMemory(&Present_Parameters, sizeof(Present_Parameters));

   // Create the Direct3D object to get everything started.
   Direct3D_Object = Direct3DCreate9(D3D_SDK_VERSION);

   // Error checking.  Make sure that it was successful.
   if(Direct3D_Object == NULL)
      {
         MessageBox(NULL, "Error, couldn't initialize DirectX!?!",
                    "Error!", MB_OK);
         return false;
      }

   // This function will get the display mode of the device and place it in DisplayMode.
   if(FAILED(Direct3D_Object->GetAdapterDisplayMode(D3DADAPTER_DEFAULT, &DisplayMode)))
      {
         MessageBox(NULL, "Error setting the display mode.", "Error!", MB_OK);
         return false;
      }

   // Get the capabilities of the hardware.
	if(FAILED(Direct3D_Object->GetDeviceCaps(D3DADAPTER_DEFAULT, D3DDEVTYPE_HAL, &D3DCaps)))
      return false;

   // Test which is supported, hardware or software vertex processing.
	DWORD VertexProcessing = 0;

	if(D3DCaps.VertexProcessingCaps != 0)
		VertexProcessing |= D3DCREATE_HARDWARE_VERTEXPROCESSING;
	else
		VertexProcessing |= D3DCREATE_SOFTWARE_VERTEXPROCESSING;

   // Here we are setting the applications parameters...
   if(fullscreen)
      {
         Present_Parameters.Windowed = FALSE;               // Window mode (fullscreen).
         Present_Parameters.BackBufferWidth = 640;
         Present_Parameters.BackBufferHeight = 480;
      }
   else
      Present_Parameters.Windowed = TRUE;                   // Window mode (not fullscreen).
   Present_Parameters.SwapEffect = D3DSWAPEFFECT_DISCARD;   // Dealing with animation (see doc).
   Present_Parameters.BackBufferFormat = DisplayMode.Format;// Render to the area of the screen.
   Present_Parameters.BackBufferCount = 1;                  // Number of back buffers.
   Present_Parameters.EnableAutoDepthStencil = TRUE;        // Check documentation.
   Present_Parameters.AutoDepthStencilFormat = D3DFMT_D16;  // Check documentation.

   // Now we must create the rendering device.
   if(FAILED(Direct3D_Object->CreateDevice(D3DADAPTER_DEFAULT, D3DDEVTYPE_HAL, hwnd,
                                           VertexProcessing,
                                           &Present_Parameters, &D3D_Device)))
      {
         MessageBox(NULL, "CreateDevice() failed!  Make sure you have DirectX 9.",
                    "Error!", MB_OK);
         return false;
      }

   // One last check to be sure.
   if(D3D_Device == NULL)
      {
         MessageBox(NULL, "D3D_Device is equal to NULL!?!", "Error!", MB_OK);
         return false;
      }

   D3D_Device->SetRenderState(D3DRS_LIGHTING, FALSE);       // Lighting off.
	D3D_Device->SetRenderState(D3DRS_CULLMODE, D3DCULL_NONE);// Culling off.
   D3D_Device->SetRenderState(D3DRS_ZENABLE, TRUE);         // Enable depth testing.

   // Here we set up the object this demo will render (for example a polygon).
   // This will load the data into the vertex buffer.
   if(!InitializeObject()) return false;

   // Set the image states to get a good quality image.
	D3D_Device->SetSamplerState(0, D3DSAMP_MINFILTER, D3DTEXF_LINEAR);
	D3D_Device->SetSamplerState(0, D3DSAMP_MAGFILTER, D3DTEXF_LINEAR);

   return true;
}


bool InitializeObject()
{
   // Square data for its position and texture coords.
	Vertex Polygon_Data[4] =
	{
		{-0.3f, 0.5f, 0.0f, 0.0, 0.0},
      {0.3f, 0.5f, 0.0f, 1.0, 0.0},
	   {-0.3f, -0.5f, 0.0f, 0.0, 1.0},
      {0.3f, -0.5f, 0.0f, 1.0, 1.0}
	};

   // Create the vertex buffer that will hold the square.
	if(FAILED(D3D_Device->CreateVertexBuffer(sizeof(Polygon_Data),
                                            0, D3DFVF_D3DVertex,
                                            D3DPOOL_DEFAULT, &Vertex_Buffer, NULL)))
      {
         return false;
      }
	
   // Pointer to the buffer.
   Vertex* Vertices;

   // Lock the buffer we can write to it.
   if(FAILED(Vertex_Buffer->Lock(0, sizeof(Polygon_Data), (void**)&Vertices, 0)))
      return false;

   // Here we copy the square's data into the vertex buffer.
   memcpy(Vertices, Polygon_Data, sizeof(Polygon_Data));

   // Unlock when your done coping data into the buffer.
   Vertex_Buffer->Unlock();

   // Here we load the texture into the texture object.  D3DXCreateTextureFromFile needs
   // our device object, the image file, and the address to our texture object.
   // With this you can load the image formats .bmp, .dds, .dib, .jpg, .png, and .tga.
   if(D3DXCreateTextureFromFile(D3D_Device, "ugp.tga", &Texture) != D3D_OK)
      return false;

   return true;
}


void RenderScene()
{
   // Clear the screen to black.
   D3D_Device->Clear(0, NULL, D3DCLEAR_TARGET | D3DCLEAR_ZBUFFER,
                     D3DCOLOR_XRGB(0,0,0), 1.0f, 0);
    
   // Tell Direct 3D we are now going to start drawing.
   D3D_Device->BeginScene();

      // SetTexture will add the image in the Texture object to all
      // things draw after this.  To Turn if off send SetTexture(0, NULL);
      D3D_Device->SetTexture(0, Texture);

	   // This will bind the vertex data in the buffer to the Direct3D device.
      D3D_Device->SetStreamSource(0, Vertex_Buffer, 0, sizeof(Vertex));

      // Set the vertex stream declaration.
      D3D_Device->SetFVF(D3DFVF_D3DVertex);

      // This will draw everything in the buffer.
      D3D_Device->DrawPrimitive(D3DPT_TRIANGLESTRIP, 0, 2);
    
   // We are done drawing for this scene.
   D3D_Device->EndScene();
    
   // Swap the old frame with the new one.
   D3D_Device->Present(NULL, NULL, NULL, NULL);
}


void ShutdownDirect3D()
{
   // Here we release the Direct3D objects.
   if(Texture != NULL) 
      {
         Texture->Release();
         Texture = NULL;
      }

   if(D3D_Device != NULL)
      {
         D3D_Device->Release();
         D3D_Device = NULL;
      }

   if(Direct3D_Object != NULL)
      {
         Direct3D_Object->Release();
         Direct3D_Object = NULL;
      }

   if(Vertex_Buffer != NULL)
      {
         // Release the buffer.
	      Vertex_Buffer->Release();
         Vertex_Buffer = NULL;
      }
}


// Recap:

// And there we have it.  With a few function calls we can have texture mapping
// in our Direct3D scene.  When your done you must remember to release the texture
// object to prevent any kind of memory leaks.


// Copyright September 2004
// All Rights Reserved!
// Allen Sherrod
// ProgrammingAce@UltimateGameProgramming.com
// www.UltimateGameProgramming.com